module Main where

import System.IO
import System.Environment
import Data.Char

import qualified Data.Array.Parallel.Prelude.Word8 as W
import GHC.PArr

import Types
import Trie
import SpellCheckUnlifted
import SpellCheckDPH

c2w :: Char -> W.Word8
c2w c = fromIntegral (ord c)
w2c :: W.Word8 -> Char
w2c w = chr (fromIntegral w)
wArr2Str :: [W.Word8] -> String
wArr2Str wArr = map w2c wArr

t2P :: Trie Char -> TrieP
t2P (Leaf c) = LeafP (c2w c)
t2P (Node c d) = NodeP (c2w c) (dic2P d)

dic2P :: Dictionary -> DictionaryP
dic2P ds = toP $ map t2P ds

doc2P :: Document -> DocumentP
doc2P doc = let doc' = filter (\s -> not $ s=="") doc
              in toP $ map (toP.(\s -> map c2w s)) doc'

-- Parse a string and return a list of words
-- (words are consecutive alphabetic characters or ' 
--  and - characters)
parseWords :: String -> String -> [String]
parseWords [] s = [s]
parseWords (x:xs) s =
    if isAlpha x || x == '\'' || x == '-' then
       parseWords xs (s++[x])
       else if s == [] then parseWords xs []
                       else (s:parseWords xs [])

-- Perform some action with each line in a file
lineMap :: Handle -> Int -> (String -> a -> a) -> a -> IO (a,Int)
lineMap h i f a = do
    eof <- hIsEOF h
    if eof then do return (a,i)
           else do l <- hGetLine h
                   lineMap h (i+1) f (f l a)

-- Read dict file and produce Trie
getDictionary :: String -> IO Dictionary
getDictionary s =
  do h <- openFile s ReadMode
     (t,i) <- lineMap h 0 (insertTrie . (map toLower)) ([]::Dictionary)
     putStrLn ("Read " ++ (show i) ++ " words from dict file " ++ s)
     return t

getDocument :: String -> IO Document
getDocument s =
  do h <- openFile s ReadMode
     eof <- hIsEOF h
     if eof then return []
            else do s <- hGetContents h
                    return (parseWords s "")

main =
 do
  args <- getArgs
  if length args == 3
     then do [x,y,z] <- getArgs
             putStrLn "Loading dictionary into Trie..."
             dic <- getDictionary x
             putStrLn "Loading Document..."
             doc <- getDocument y
             if z == "0"
                then do
                     let spellErrs = spellCheckU dic doc
                     putStrLn "Performing unlifted parallel spell check..."
                     putStrLn $ show (length spellErrs ) ++ " errors found!"
                     putStrLn "Done!"
                else do
                     putStrLn "Performing nested parallel spell check..."
                     let spellErrs = map wArr2Str $ fromP $ mapP fromP (spellCheckP (dic2P dic) (doc2P doc))
                     putStrLn $ show (length spellErrs ) ++ " errors found!"
                     putStrLn "Done!"
     else do putStrLn "Usage: SpellCheck DIC_FILE DOC_FILE (0|1)"
             putStrLn "0=Unlifted, 1=Nested"
             putStrLn "E.g. SpellCheck /usr/share/dict/words ~/doc.txt 0"